// Stealth AI Text Extractor - Firefox Version - Automatic Mode with SPA/AJAX Support

class StealthAIExtractor {
  constructor() {
    this.apiKey = "AIzaSyAipHAMwTdsZ3Bqti85L_LUN2Xd_erfN6g";
    this.isRunning = false;
    this.extractionInterval = null;
    this.mutationObserver = null; // For SPA/AJAX support
    this.questionCounter = 0;
    this.answerDisplay = null;
    this.isDisplayVisible = false;
    this.lastAnswer = "";
    this.lastProcessedText = ""; // Track last processed text to avoid duplicates
    this.processedElements = new Set(); // Move to instance level and clear periodically
    this.isProcessing = false; // Prevent concurrent processing
    this.errorCount = 0;

    // Firefox compatibility check
    this.isFirefox =
      typeof browser !== "undefined" || navigator.userAgent.includes("Firefox");

    this.init();
  }

  async init() {
    this.createAnswerDisplay();
    this.setupClickListener();
    // Clear any previous session data
    this.clearSessionData();
    // Wait 3 seconds after page load, then start automatic extraction
    setTimeout(() => {
      this.startAutomaticExtraction();
      this.setupMutationObserver(); // Add SPA/AJAX support
    }, 3000);
  }

  setupMutationObserver() {
    // Detect DOM changes for SPA/AJAX content updates
    this.mutationObserver = new MutationObserver((mutations) => {
      let hasSignificantChanges = false;

      mutations.forEach((mutation) => {
        // Check for added nodes with text content
        if (mutation.type === "childList") {
          mutation.addedNodes.forEach((node) => {
            if (node.nodeType === Node.ELEMENT_NODE) {
              const textContent = node.textContent || "";
              // Look for question patterns or significant text changes
              if (
                textContent.length > 50 &&
                (textContent.includes("?") ||
                  /[A-E][\)\.]/.test(textContent) ||
                  textContent.length > 200)
              ) {
                hasSignificantChanges = true;
              }
            }
          });
        }

        // Check for text content changes
        if (
          mutation.type === "characterData" &&
          mutation.target.textContent.length > 20
        ) {
          hasSignificantChanges = true;
        }
      });

      // If significant changes detected, analyze after short delay
      if (hasSignificantChanges && !this.isProcessing) {
        console.log(
          "Stealth AI Extractor (Firefox): DOM change detected (SPA/AJAX)"
        );
        setTimeout(() => {
          this.processedElements.clear(); // Clear cache for new content
          this.extractAndAnalyze();
        }, 1000); // Short delay to let content fully load
      }
    });

    // Start observing the document with the configured parameters
    this.mutationObserver.observe(document.body, {
      childList: true,
      subtree: true,
      characterData: true,
    });

    console.log(
      "Stealth AI Extractor (Firefox): SPA/AJAX monitoring activated"
    );
  }

  clearSessionData() {
    // Clear previous session data to avoid conflicts
    sessionStorage.removeItem("aiExtractorLastResult");
    console.log("Stealth AI Extractor (Firefox): Session data cleared");
  }

  createAnswerDisplay() {
    // Create minimal answer display in top-right corner
    this.answerDisplay = document.createElement("div");
    this.answerDisplay.id = "stealth-answer-display";
    this.answerDisplay.className = "hidden";
    this.answerDisplay.textContent = "Right-click to show";
    document.body.appendChild(this.answerDisplay);
  }

  setupClickListener() {
    // Right click anywhere on page to toggle answer display
    document.addEventListener("contextmenu", (e) => {
      // Only trigger if not right-clicking on other interactive elements
      if (!e.target.closest("button, a, input, select, textarea, img")) {
        e.preventDefault(); // Prevent default context menu
        this.toggleAnswerDisplay();
      }
    });

    // Right click on the answer display itself to toggle
    this.answerDisplay.addEventListener("contextmenu", (e) => {
      e.preventDefault(); // Prevent default context menu
      e.stopPropagation();
      this.toggleAnswerDisplay();
    });
  }

  toggleAnswerDisplay() {
    if (this.isDisplayVisible) {
      this.hideAnswerDisplay();
    } else {
      this.showAnswerDisplay();
    }
  }

  showAnswerDisplay() {
    this.isDisplayVisible = true;
    this.answerDisplay.className = "visible";
    if (this.lastAnswer) {
      this.answerDisplay.textContent = this.lastAnswer;
    } else {
      this.answerDisplay.textContent = "...";
    }
    console.log("Answer display shown");
  }

  hideAnswerDisplay() {
    this.isDisplayVisible = false;
    this.answerDisplay.className = "hidden";
    console.log("Answer display hidden");
  }

  startAutomaticExtraction() {
    if (this.isRunning) return;

    this.isRunning = true;
    console.log(
      "Stealth AI Extractor (Firefox): Starting automatic mode with SPA support"
    );

    // Run immediately, then every 15 seconds (longer interval since we have mutation observer)
    this.extractAndAnalyze();
    this.extractionInterval = setInterval(() => {
      // More aggressive cleanup every 10th question
      if (this.questionCounter > 0 && this.questionCounter % 10 === 0) {
        console.log(
          "Stealth AI Extractor (Firefox): Performing deep cleanup after 10 questions"
        );
        this.performDeepCleanup();
      }

      // Clear processed elements periodically to avoid memory issues
      this.processedElements.clear();
      this.extractAndAnalyze();
    }, 15000); // Increased to 15 seconds since mutation observer handles immediate changes
  }

  performDeepCleanup() {
    // Clear all caches and force garbage collection
    this.processedElements.clear();
    this.lastProcessedText = "";

    // Clear old session storage entries
    const oldEntries = Object.keys(sessionStorage).filter(
      (key) => key.startsWith("aiExtractor") && key !== "aiExtractorLastResult"
    );
    oldEntries.forEach((key) => sessionStorage.removeItem(key));

    // Firefox specific memory management
    if (this.isFirefox && typeof Components !== "undefined") {
      try {
        Components.utils.forceGC();
      } catch (e) {
        // Fallback for newer Firefox versions
        if (window.gc) {
          window.gc();
        }
      }
    } else if (window.gc) {
      window.gc();
    }

    console.log(
      `Stealth AI Extractor (Firefox): Deep cleanup completed after ${this.questionCounter} questions`
    );
  }

  stopAutomaticExtraction() {
    if (this.extractionInterval) {
      clearInterval(this.extractionInterval);
      this.extractionInterval = null;
    }

    if (this.mutationObserver) {
      this.mutationObserver.disconnect();
      this.mutationObserver = null;
    }

    this.isRunning = false;
    // Clear memory
    this.processedElements.clear();
    this.lastProcessedText = "";
    console.log("Stealth AI Extractor (Firefox): Stopped");
  }

  async extractAndAnalyze() {
    if (this.isProcessing) {
      console.log(
        "Stealth AI Extractor (Firefox): Already processing, skipping..."
      );
      return;
    }

    try {
      this.isProcessing = true;
      const extractedText = this.extractWithDOM();

      if (!extractedText.trim() || extractedText.length < 20) {
        console.log(
          "Stealth AI Extractor (Firefox): Insufficient text content"
        );
        return;
      }

      // Check similarity - be extremely lenient for exam questions
      if (this.lastProcessedText) {
        const similarity = this.calculateSimilarity(
          extractedText,
          this.lastProcessedText
        );

        // Only block if content is >98% identical (almost exact duplicates)
        if (similarity > 0.98) {
          console.log(
            "Stealth AI Extractor (Firefox): Content nearly identical (>98%), skipping..."
          );
          return;
        }
      }

      this.lastProcessedText = extractedText;

      console.log(
        "Stealth AI Extractor (Firefox): Analyzing text...",
        extractedText.length,
        "characters"
      );

      const response = await this.sendToGemini(extractedText);

      // Log the result silently (stealth mode)
      console.log("Stealth AI Extractor (Firefox): Analysis result:", response);

      // Update answer based on response
      if (response && response !== "NO_QUESTION") {
        this.updateAnswer(response);
      } else {
        // No question found, show question mark
        this.lastAnswer = "?";
        if (this.isDisplayVisible) {
          this.answerDisplay.textContent = "?";
        }
      }

      // Store result in session storage for potential retrieval
      sessionStorage.setItem(
        "aiExtractorLastResult",
        JSON.stringify({
          timestamp: Date.now(),
          result: response,
          textLength: extractedText.length,
          questionNumber: this.questionCounter,
        })
      );

      // Reset error counter on successful operation
      this.errorCount = 0;
    } catch (error) {
      console.log("Stealth AI Extractor (Firefox): Error:", error.message);
      this.handleError(error);
    } finally {
      this.isProcessing = false;
    }
  }

  calculateSimilarity(text1, text2) {
    // Improved similarity calculation for better question detection
    const len1 = text1.length;
    const len2 = text2.length;

    if (len1 === 0 && len2 === 0) return 1.0;
    if (len1 === 0 || len2 === 0) return 0.0;

    // Compare first 1500 characters for better accuracy
    const sample1 = text1.substring(0, 1500).toLowerCase().replace(/\s+/g, " ");
    const sample2 = text2.substring(0, 1500).toLowerCase().replace(/\s+/g, " ");

    // Check for question patterns - if different question structures, reduce similarity
    const questionWords1 = (
      sample1.match(
        /\b(what|which|who|where|when|how|why|is|are|does|do|can|will|would|should)\b/g
      ) || []
    ).length;
    const questionWords2 = (
      sample2.match(
        /\b(what|which|who|where|when|how|why|is|are|does|do|can|will|would|should)\b/g
      ) || []
    ).length;

    const questionMarks1 = (sample1.match(/\?/g) || []).length;
    const questionMarks2 = (sample2.match(/\?/g) || []).length;

    // Check for multiple choice patterns (A) B) C) etc.
    const choicePatterns1 = (sample1.match(/[a-e][\)\.\:]|\([a-e]\)/g) || [])
      .length;
    const choicePatterns2 = (sample2.match(/[a-e][\)\.\:]|\([a-e]\)/g) || [])
      .length;

    // Check for question numbers (1. 2. 3. etc.)
    const questionNumbers1 = (
      sample1.match(/\b\d+\.\s*soru|\b\d+\s*[\)\.]|\bsoru\s*\d+/g) || []
    ).length;
    const questionNumbers2 = (
      sample2.match(/\b\d+\.\s*soru|\b\d+\s*[\)\.]|\bsoru\s*\d+/g) || []
    ).length;

    // If question structures are significantly different, lower similarity
    const questionDiff =
      Math.abs(questionWords1 - questionWords2) +
      Math.abs(questionMarks1 - questionMarks2);
    const choiceDiff = Math.abs(choicePatterns1 - choicePatterns2);
    const numberDiff = Math.abs(questionNumbers1 - questionNumbers2);

    // Calculate Jaccard similarity with character bigrams
    const bigrams1 = this.getBigrams(sample1);
    const bigrams2 = this.getBigrams(sample2);

    const union = new Set([...bigrams1, ...bigrams2]);
    const intersection = new Set([...bigrams1].filter((x) => bigrams2.has(x)));

    let similarity = intersection.size / union.size;

    // Reduce similarity if question structures are different
    if (questionDiff > 2) {
      similarity *= 0.7; // Reduce by 30%
    }
    if (choiceDiff > 3) {
      similarity *= 0.8; // Reduce by 20% for different choice patterns
    }
    if (numberDiff > 0) {
      similarity *= 0.5; // Reduce by 50% if different question numbers detected
    }

    // Extract key content words (longer than 4 characters) and compare
    const contentWords1 = new Set(sample1.match(/\b\w{5,}\b/g) || []);
    const contentWords2 = new Set(sample2.match(/\b\w{5,}\b/g) || []);

    const wordUnion = new Set([...contentWords1, ...contentWords2]);
    const wordIntersection = new Set(
      [...contentWords1].filter((x) => contentWords2.has(x))
    );

    const wordSimilarity =
      wordUnion.size > 0 ? wordIntersection.size / wordUnion.size : 0;

    // Combine bigram and word similarity (give more weight to content words)
    const finalSimilarity = similarity * 0.6 + wordSimilarity * 0.4;

    // Special handling for short texts (exam questions) - be more lenient
    if (Math.min(len1, len2) < 200) {
      // For very short texts (under 100 chars), be extremely lenient
      if (Math.min(len1, len2) < 100) {
        return Math.min(finalSimilarity * 0.1, 1.0);
      }
      // For short texts, reduce calculated similarity by 70% to be more flexible
      return Math.min(finalSimilarity * 0.3, 1.0);
    }

    return Math.min(finalSimilarity, 1.0);
  }

  getBigrams(text) {
    const bigrams = new Set();
    for (let i = 0; i < text.length - 1; i++) {
      bigrams.add(text.substring(i, i + 2));
    }
    return bigrams;
  }

  handleError(error) {
    // Track consecutive errors
    if (!this.errorCount) this.errorCount = 0;
    this.errorCount++;

    console.log(
      `Stealth AI Extractor (Firefox): Error #${this.errorCount}: ${error.message}`
    );

    // Reset on error to allow retry
    this.lastProcessedText = "";

    // If too many consecutive errors, restart the extension
    if (this.errorCount >= 3) {
      console.log(
        "Stealth AI Extractor (Firefox): Too many errors, restarting..."
      );
      this.restart();
    }
  }

  restart() {
    console.log("Stealth AI Extractor (Firefox): Performing restart...");

    // Stop current operation
    this.stopAutomaticExtraction();

    // Reset all state
    this.errorCount = 0;
    this.questionCounter = 0;
    this.lastAnswer = "";
    this.lastProcessedText = "";
    this.processedElements.clear();

    // Clear session storage
    this.clearSessionData();

    // Restart after short delay
    setTimeout(() => {
      this.startAutomaticExtraction();
      this.setupMutationObserver();
      console.log("Stealth AI Extractor (Firefox): Restart completed");
    }, 2000);
  }

  updateAnswer(answer) {
    // Check if it's a multiple choice answer (A, B, C, D, E)
    const multipleChoicePattern = /^[A-E]$/;
    const isMultipleChoice = multipleChoicePattern.test(answer.trim());

    if (isMultipleChoice) {
      this.questionCounter++;
      this.lastAnswer = `${this.questionCounter}. ${answer.trim()}`;
      console.log(`Question ${this.questionCounter}: Answer ${answer.trim()}`);
    } else {
      // For non-multiple choice answers, show briefly without counter
      this.lastAnswer =
        answer.substring(0, 30) + (answer.length > 30 ? "..." : "");
    }

    // Update display if currently visible
    if (this.isDisplayVisible) {
      this.answerDisplay.textContent = this.lastAnswer;
    }
  }

  extractWithDOM() {
    const elementsToExtract = [
      "p",
      "h1",
      "h2",
      "h3",
      "h4",
      "h5",
      "h6",
      "li",
      "td",
      "th",
      "span",
      "div",
      "article",
      "section",
      "main",
      "blockquote",
    ];

    let text = "";
    const excludeKeywords = [
      "cookie",
      "privacy",
      "terms",
      "footer",
      "header",
      "navigation",
      "menu",
    ];

    elementsToExtract.forEach((tag) => {
      const elements = document.querySelectorAll(tag);
      elements.forEach((el) => {
        // Skip if already processed in current extraction
        if (this.processedElements.has(el)) return;

        // Check if element is visible and has meaningful content
        if (
          this.isElementVisible(el) &&
          el.innerText &&
          el.innerText.trim().length > 3
        ) {
          const elementText = el.innerText.trim();
          const elementClass = el.className.toLowerCase();
          const elementId = el.id.toLowerCase();

          // Skip elements that are likely navigation, ads, or boilerplate
          const shouldSkip = excludeKeywords.some(
            (keyword) =>
              elementClass.includes(keyword) ||
              elementId.includes(keyword) ||
              elementText.toLowerCase().includes("cookie policy") ||
              elementText.toLowerCase().includes("terms of service")
          );

          if (
            !shouldSkip &&
            elementText.length > 5 &&
            !text.includes(elementText.substring(0, 50))
          ) {
            text += elementText + " ";
            this.processedElements.add(el);
          }
        }
      });
    });

    // Clean and limit text
    const cleanedText = text.replace(/\s+/g, " ").trim();
    return cleanedText.substring(0, 6000);
  }

  isElementVisible(el) {
    const style = window.getComputedStyle(el);
    return (
      el.offsetParent !== null &&
      style.display !== "none" &&
      style.visibility !== "hidden" &&
      style.opacity !== "0" &&
      el.offsetWidth > 0 &&
      el.offsetHeight > 0
    );
  }

  async sendToGemini(text) {
    const prompt = `Analyze the following text. If this contains a question or quiz, provide a direct answer. If it's multiple choice (A, B, C, D, E), respond with just the letter. For other questions, provide a brief, direct answer. If no clear question is found, respond with "NO_QUESTION".

Text to analyze:
${text}

Response format:
- For multiple choice: Just the letter (A, B, C, D, or E)
- For other questions: Brief direct answer
- For non-questions: NO_QUESTION`;

    // Add timeout to prevent hanging requests - Firefox needs longer timeout
    const controller = new AbortController();
    const timeoutId = setTimeout(
      () => controller.abort(),
      this.isFirefox ? 20000 : 15000
    );

    try {
      // Firefox specific fetch options
      const fetchOptions = {
        method: "POST",
        headers: {
          "Content-Type": "application/json",
        },
        body: JSON.stringify({
          contents: [
            {
              parts: [
                {
                  text: prompt,
                },
              ],
            },
          ],
        }),
        signal: controller.signal,
      };

      // Add Firefox specific options
      if (this.isFirefox) {
        fetchOptions.credentials = "omit";
        fetchOptions.cache = "no-cache";
      }

      const response = await fetch(
        `https://generativelanguage.googleapis.com/v1beta/models/gemini-2.5-pro-preview-05-06:generateContent?key=${this.apiKey}`,
        fetchOptions
      );

      clearTimeout(timeoutId);

      if (!response.ok) {
        throw new Error("Gemini API error: " + response.statusText);
      }

      const data = await response.json();

      if (data.candidates && data.candidates[0] && data.candidates[0].content) {
        return data.candidates[0].content.parts[0].text.trim();
      } else {
        throw new Error("Invalid response from Gemini API");
      }
    } catch (error) {
      clearTimeout(timeoutId);
      if (error.name === "AbortError") {
        throw new Error("Request timeout");
      }

      // Firefox specific error handling
      if (this.isFirefox && error.message.includes("NetworkError")) {
        throw new Error("Firefox network error - check CORS settings");
      }

      throw error;
    }
  }
}

// Firefox-compatible auto-start when page loads
if (document.readyState === "loading") {
  document.addEventListener("DOMContentLoaded", () => {
    // Clean up any existing instance
    if (window.stealthExtractor) {
      window.stealthExtractor.stopAutomaticExtraction();
    }
    window.stealthExtractor = new StealthAIExtractor();
  });
} else {
  // Clean up any existing instance
  if (window.stealthExtractor) {
    window.stealthExtractor.stopAutomaticExtraction();
  }
  window.stealthExtractor = new StealthAIExtractor();
}

// Clean up on page unload
window.addEventListener("beforeunload", () => {
  if (window.stealthExtractor) {
    window.stealthExtractor.stopAutomaticExtraction();
  }
});

// Firefox specific event handling
if (typeof browser !== "undefined") {
  // Additional Firefox-specific setup if needed
  console.log(
    "Stealth AI Extractor: Firefox WebExtension environment detected"
  );
}
